package checkers.nonnull;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.*;

import checkers.quals.NonNull;
import checkers.types.*;

import com.sun.source.tree.*;
import com.sun.tools.javac.tree.*;

/**
 * Adds support for the {@code @NonNull} type annotation to 
 * {@link AnnotatedTypeFactory}. This means that the 
 * {@link NonnullAnnotatedTypeFactory#getClass} and
 * {@link NonnullAnnotatedTypeFactory#getMethod} methods will regard some 
 * inputs as {@code @NonNull} even if they do not carry a {@code @NonNull} 
 * annotation. These include:
 *
 * <ul>
 *  <li>primitive, class, and enum types
 *  <li>{@code super} and {@code this}
 *  <li>the results of binary operations
 *  <li>literals (except the null literal) and arrays of literals
 * </ul>
 */
public class NonnullAnnotatedTypeFactory extends AnnotatedTypeFactory {

    /**
     * Creates an annotated type factory that treats certain inputs as having
     * {@code @NonNull} annotations even if they do not explicitly carry a 
     * {@code @NonNull} annotation.
     *
     * @param env the {@link ProcessingEnvironment} to use for tree and type
     *        utilities
     * @param root the root of the syntax tree (used for TreePaths)
     */
    public NonnullAnnotatedTypeFactory(ProcessingEnvironment env, CompilationUnitTree root) {
        super(env, root);
    }

    @Override
    public AnnotatedClassType createClassType() {
        return new NonnullAnnotatedClassType();
    }
    
    @Override
    public AnnotatedClassType getClass(Tree tree) {
    
        // If the given tree is a method invocation, return the annotated type
        // of the method's return value.
        if (tree instanceof MethodInvocationTree)
            return getMethod(tree).getAnnotatedReturnType();
        
        // Get the original annotated type, to which @NonNull annotations
        // may be added by the remainder of this method.
        AnnotatedClassType type = super.getClass(tree);
        
        tree = TreeInfo.skipParens((JCTree)tree);

        // FIXME *** for flow sensitive analysis ("if (x != null)")
        /*
        TreePath path = TreePath.getPath(root, tree);
        Tree child = null;
        while (path != null && path.getParentPath() != null) {
            Tree leaf = path.getLeaf();
            if (leaf.getKind() == Tree.Kind.IF) {
                IfTree ift = (IfTree)tree;
                ift.getCondition()
            }
            child = path.getLeaf();
            path = path.getParentPath();
        }
        */
        
        if (tree instanceof TypeCastTree) {
            
            TypeCastTree tc = (TypeCastTree)tree;
            if (tc.getType().getKind() == Tree.Kind.PRIMITIVE_TYPE)
                type.include(NonNull.class);
            
        } else if (tree instanceof IdentifierTree) {
            
            IdentifierTree id = (IdentifierTree)tree;
            if (id.getName().contentEquals("super") || id.getName().contentEquals("this"))
                type.include(NonNull.class);
            
            Element elt = InternalUtils.symbol(id);
            if (elt.getKind() == ElementKind.CLASS || elt.getKind() == ElementKind.ENUM)
                type.include(NonNull.class);
            
        } else if (tree instanceof BinaryTree) {
            type.include(NonNull.class);
        }
        
        Element typeElt = type.getElement();
        if (typeElt instanceof VariableElement) {
            if (typeElt.asType().getKind().isPrimitive())
                type.include(NonNull.class);
        }
                
        switch (tree.getKind()) {
        case NEW_ARRAY:
            NewArrayTree nat = (NewArrayTree)tree;
            boolean allNN = true;
            for (ExpressionTree et : nat.getInitializers()) {
                if (!getClass(et).isAnnotatedWith(NonNull.class)) {
                    allNN = false;
                    break;
                }
            }
            if (allNN)
                type.annotateAt(AnnotationLocation.fromArray(new int[] { 0 }), NonNull.class);
            // Don't break.
        case NEW_CLASS:
        case BOOLEAN_LITERAL:
        case DOUBLE_LITERAL:
        case FLOAT_LITERAL:
        case INT_LITERAL:
        case LONG_LITERAL:
        case STRING_LITERAL:
        case CONDITIONAL_AND:
        case CONDITIONAL_OR:
        // TODO *** test this!
        //case CONDITIONAL_EXPRESSION: 
            type.include(NonNull.class);
        }
       
        return type;
    }

    @Override
    protected AnnotatedClassType returnType(ExecutableElement method) {
        AnnotatedClassType type = super.returnType(method);
        
        if (method.getReturnType().getKind().isPrimitive())
            type.include(NonNull.class);
        
        return type;
    }

    @Override
    protected AnnotatedClassType receiverType(ExecutableElement method) {
        AnnotatedClassType type = super.receiverType(method);
        type.include(NonNull.class);
        return type;
    }

    @Override
    public AnnotatedMethodType getMethod(Tree tree) {
        AnnotatedMethodType type = super.getMethod(tree);
        
        // TODO *** remove this eventually and replace with a @SuppressWarnings mechanism
        
        // Make static final fields implicitly @Nonnull (so we don't get type 
        // errors from things like System.out)
        if (tree instanceof MethodInvocationTree) {
            MethodInvocationTree mi = (MethodInvocationTree)tree;
            if (mi.getMethodSelect() instanceof MemberSelectTree) {
                MemberSelectTree ms = (MemberSelectTree)mi.getMethodSelect();
                Element elt = InternalUtils.symbol(ms.getExpression());
                if (elt != null /*&& elt.getModifiers().contains(Modifier.FINAL)*/
                        && elt.getModifiers().contains(Modifier.STATIC))
                    type.getAnnotatedReceiverType().exclude(NonNull.class);
                if (super.isOuterClassReference(ms))
                    type.getAnnotatedReturnType().include(NonNull.class);
            }
        }
        
        return type;
    }
}
